/* 
 * Copyright (c) 2010 ISPRAS
 *
 * Institute for System Programming of the Russian Academy of Sciences (ISPRAS)
 * 25 A.Solzhenitsyn st. Moscow 109004 Russia
 * http://hardware.ispras.ru
 *
 * All rights reserved.
 *
 * This source code is free software; you can redistribute it
 * and/or modify it in source code form under the terms of the GNU
 * General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 * $Id: testbench.c,v 1.3 2010/03/11 11:54:58 kamkin Exp $
 */

#include <assert.h>
#include <stdio.h>

#include "common.h"
#include "testbench.h"

#include "../error.h"

static void vt_print_input_testbench_v(FILE *file, vt_wire_t *wire, vt_options_t *options)
{
    size_t width;

    assert(wire != NULL);

    width = wire->max - wire->min + 1;

    if(wire->input != VT_DATA)
    {
        if(width > 1)
            { fprintf(file, "    reg [%d:%d] %s;\n", wire->max, wire->min, wire->name); }
		else
            { fprintf(file, "    reg %s;\n", wire->name); }
	}
    else
    {
        if(width > 1)
        {
            fprintf(file, "    reg [%d:%d] %s;\n", wire->max, wire->min, wire->name);
            fprintf(file, "    reg [%d:%d] %s_reg;\n", wire->max, wire->min, wire->name);
        }
        else
        {
            fprintf(file, "    reg %s;\n", wire->name);
            fprintf(file, "    reg %s_reg;\n", wire->name);
        }
    }
}

static void vt_print_output_testbench_v(FILE *file, vt_wire_t *wire, vt_options_t *options)
{
    size_t width;

    assert(wire != NULL);

    width = wire->max - wire->min + 1;

    if(width > 1)
        { fprintf(file, "    wire [%d:%d] %s;\n", wire->max, wire->min, wire->name); }
    else
        { fprintf(file, "    wire %s;\n", wire->name); }
}

static void vt_print_reset_testbench_v(FILE *file, vt_wire_t *wire, vt_options_t *options)
{
    int width, i;

    assert(wire != NULL);

    width = wire->max - wire->min + 1;

    fprintf(file, "        %s = %d'h", wire->name, width);

    for(i = 0; i < width >> 2; i++)
        { fprintf(file, "0"); }

    if((width & 0x7) != 0)
    	{ fprintf(file, "0"); }

    fprintf(file, ";\n");
}

static void vt_print_clock_cycle(FILE *file, vt_options_t *options)
{
    fprintf(file, "%s = 0; #(delay_cycle); %s = 1; #(delay_cycle);\n", options->clock_input, options->clock_input);
}

void vt_print_testbench_v(FILE *file, vt_design_t *design, vt_options_t *options)
{
    vt_item_t *item;

    assert(file != NULL);
    assert(design != NULL);
    assert(options != NULL);

    fprintf(file, "`timescale 1 ns/1 ps\n\n");
    
    fprintf(file, "module testbench();\n\n");
    fprintf(file, "    parameter delay_cycle = 5;\n\n");

    for(item = design->inputs->head; item; item = item->next)
        { vt_print_input_testbench_v(file, item->wire, options); }

    fprintf(file, "    reg run;\n");

    for(item = design->outputs->head; item; item = item->next)
        { vt_print_output_testbench_v(file, item->wire, options); }

    vt_print_newline(file);

    fprintf(file, "    always @(posedge %s) begin\n", options->clock_input);

    for(item = design->inputs->head; item; item = item->next)
    {
        vt_wire_t *wire = item->wire;

        if(wire->input == VT_DATA)
            { fprintf(file, "        %s_reg <= %s;\n", wire->name, wire->name); }
    }

    fprintf(file, "    end\n\n");

    fprintf(file, "    %s target (\n", design->name);

    for(item = design->inputs->head; item; item = item->next)
    {
        vt_wire_t *wire = item->wire;

        if(wire->input == VT_DATA)
            { fprintf(file, "        .%s(%s_reg)", wire->name, wire->name); }
        else
            { fprintf(file, "        .%s(%s)", wire->name, wire->name); }
            
        if(item->next || design->outputs->head)
            { fprintf(file, ",\n"); }
    }

    for(item = design->outputs->head; item; item = item->next)
    {
        vt_wire_t *wire = item->wire;

        fprintf(file, "        .%s(%s)", wire->name, wire->name);

        if(item->next)
            { fprintf(file, ",\n"); }
    }

    fprintf(file, "\n    );\n\n");

    fprintf(file, "    initial begin\n");
    fprintf(file, "        $dumpfile(\"testbench.vcd\");\n");
    fprintf(file, "        $dumpvars(1, testbench);\n");

    for(item = design->inputs->head; item; item = item->next)
        { vt_print_reset_testbench_v(file, item->wire, options); }

    fprintf(file, "        %s = 1'h0;\n", options->reset_input);

    vt_print_indent(file, 2); vt_print_clock_cycle(file, options);
    vt_print_indent(file, 2); vt_print_clock_cycle(file, options);
    vt_print_indent(file, 2); vt_print_clock_cycle(file, options);

    fprintf(file, "        %s = 1'h1;\n", options->reset_input);
    fprintf(file, "        $setScenario();\n");
    fprintf(file, "        run = 1'h1;\n");
    fprintf(file, "        while(run) begin\n");
    fprintf(file, "            $applyAction();\n");

    vt_print_indent(file, 3); vt_print_clock_cycle(file, options);

    fprintf(file, "            $checkAction();\n");
    fprintf(file, "        end\n");

    vt_print_indent(file, 2); vt_print_clock_cycle(file, options);

    fprintf(file, "    end\n");
    fprintf(file, "endmodule\n\n");
}
